"""Helper functions to compute hierarchical evaluation metrics."""
import numpy as np
from sklearn.utils import check_array

from hiclass.HierarchicalClassifier import make_leveled


def _validate_input(y_true, y_pred):
    assert len(y_true) == len(y_pred)
    y_pred = make_leveled(y_pred)
    y_true = make_leveled(y_true)
    y_true = check_array(y_true, dtype=None, ensure_2d=False, allow_nd=True)
    y_pred = check_array(y_pred, dtype=None, ensure_2d=False, allow_nd=True)
    return y_true, y_pred


def precision(y_true: np.ndarray, y_pred: np.ndarray, average: str = "micro"):
    r"""
    Compute hierarchical precision score.

    Parameters
    ----------
    y_true : np.array of shape (n_samples, n_levels)
        Ground truth (correct) labels.
    y_pred : np.array of shape (n_samples, n_levels)
        Predicted labels, as returned by a classifier.
    average: {"micro", "macro"}, str, default="micro"
        This parameter determines the type of averaging performed during the computation:

        - `micro`: The precision is computed by summing over all individual instances, :math:`\displaystyle{hP = \frac{\sum_{i=1}^{n}| \alpha_i \cap \beta_i |}{\sum_{i=1}^{n}| \alpha_i |}}`, where :math:`\alpha_i` is the set consisting of the most specific classes predicted for test example :math:`i` and all their ancestor classes, while :math:`\beta_i` is the set containing the true most specific classes of test example :math:`i` and all their ancestors, with summations computed over all test examples.
        - `macro`: The precision is computed for each instance and then averaged, :math:`\displaystyle{hP = \frac{\sum_{i=1}^{n}hP_{i}}{n}}`, where :math:`\alpha_i` is the set consisting of the most specific classes predicted for test example :math:`i` and all their ancestor classes, while :math:`\beta_i` is the set containing the true most specific classes of test example :math:`i` and all their ancestors.

    Returns
    -------
    precision : float
        What proportion of positive identifications was actually correct?
    """
    y_true, y_pred = _validate_input(y_true, y_pred)
    functions = {
        "micro": _precision_micro,
        "macro": _precision_macro,
    }
    return functions[average](y_true, y_pred)


def _precision_micro(y_true: np.ndarray, y_pred: np.ndarray):
    precision_micro = {
        1: _precision_micro_1d,
        2: _precision_micro_2d,
        3: _precision_micro_3d,
    }
    return precision_micro[y_true.ndim](y_true, y_pred)


def _precision_micro_1d(y_true: np.ndarray, y_pred: np.ndarray):
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        ground_truth_set = set([ground_truth])
        ground_truth_set.discard("")
        predicted_set = set([prediction])
        predicted_set.discard("")
        sum_intersection = sum_intersection + len(
            ground_truth_set.intersection(predicted_set)
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(predicted_set)
    return sum_intersection / sum_prediction_and_ancestors


def _precision_micro_2d(y_true: np.ndarray, y_pred: np.ndarray):
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        ground_truth_set = set(ground_truth)
        ground_truth_set.discard("")
        predicted_set = set(prediction)
        predicted_set.discard("")
        sum_intersection = sum_intersection + len(
            ground_truth_set.intersection(predicted_set)
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(predicted_set)
    return sum_intersection / sum_prediction_and_ancestors


def _precision_micro_3d(y_true: np.ndarray, y_pred: np.ndarray):
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for row_ground_truth, row_prediction in zip(y_true, y_pred):
        ground_truth_set = set()
        predicted_set = set()
        for ground_truth, prediction in zip(row_ground_truth, row_prediction):
            ground_truth_set.update(ground_truth)
            predicted_set.update(prediction)
        ground_truth_set.discard("")
        predicted_set.discard("")
        sum_intersection = sum_intersection + len(
            ground_truth_set.intersection(predicted_set)
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(predicted_set)
    return sum_intersection / sum_prediction_and_ancestors


def _precision_macro(y_true: np.ndarray, y_pred: np.ndarray):
    return _compute_macro(y_true, y_pred, _precision_micro)


def recall(y_true: np.ndarray, y_pred: np.ndarray, average: str = "micro"):
    r"""
    Compute hierarchical recall score.

    Parameters
    ----------
    y_true : np.array of shape (n_samples, n_levels)
        Ground truth (correct) labels.
    y_pred : np.array of shape (n_samples, n_levels)
        Predicted labels, as returned by a classifier.
    average: {"micro", "macro"}, str, default="micro"
        This parameter determines the type of averaging performed during the computation:

        - `micro`: The recall is computed by summing over all individual instances, :math:`\displaystyle{hR = \frac{\sum_{i=1}^{n}|\alpha_i \cap \beta_i|}{\sum_{i=1}^{n}|\beta_i|}}`, where :math:`\alpha_i` is the set consisting of the most specific classes predicted for test example :math:`i` and all their ancestor classes, while :math:`\beta_i` is the set containing the true most specific classes of test example :math:`i` and all their ancestors, with summations computed over all test examples.
        - `macro`: The recall is computed for each instance and then averaged, :math:`\displaystyle{hR = \frac{\sum_{i=1}^{n}hR_{i}}{n}}`, where :math:`\alpha_i` is the set consisting of the most specific classes predicted for test example :math:`i` and all their ancestor classes, while :math:`\beta_i` is the set containing the true most specific classes of test example :math:`i` and all their ancestors.

    Returns
    -------
    recall : float
        What proportion of actual positives was identified correctly?
    """
    y_true, y_pred = _validate_input(y_true, y_pred)
    functions = {
        "micro": _recall_micro,
        "macro": _recall_macro,
    }
    return functions[average](y_true, y_pred)


def _recall_micro(y_true: np.ndarray, y_pred: np.ndarray):
    recall_micro = {
        1: _recall_micro_1d,
        2: _recall_micro_2d,
        3: _recall_micro_3d,
    }
    return recall_micro[y_true.ndim](y_true, y_pred)


def _recall_micro_1d(y_true: np.ndarray, y_pred: np.ndarray):
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        ground_truth_set = set([ground_truth])
        ground_truth_set.discard("")
        predicted_set = set([prediction])
        predicted_set.discard("")
        sum_intersection = sum_intersection + len(
            ground_truth_set.intersection(predicted_set)
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(
            ground_truth_set
        )
    recall = sum_intersection / sum_prediction_and_ancestors
    return recall


def _recall_micro_2d(y_true: np.ndarray, y_pred: np.ndarray):
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        ground_truth_set = set(ground_truth)
        ground_truth_set.discard("")
        predicted_set = set(prediction)
        predicted_set.discard("")
        sum_intersection = sum_intersection + len(
            ground_truth_set.intersection(predicted_set)
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(
            ground_truth_set
        )
    recall = sum_intersection / sum_prediction_and_ancestors
    return recall


def _recall_micro_3d(y_true: np.ndarray, y_pred: np.ndarray):
    sum_intersection = 0
    sum_prediction_and_ancestors = 0
    for row_ground_truth, row_prediction in zip(y_true, y_pred):
        ground_truth_set = set()
        predicted_set = set()
        for ground_truth, prediction in zip(row_ground_truth, row_prediction):
            ground_truth_set.update(ground_truth)
            predicted_set.update(prediction)
        ground_truth_set.discard("")
        predicted_set.discard("")
        sum_intersection = sum_intersection + len(
            ground_truth_set.intersection(predicted_set)
        )
        sum_prediction_and_ancestors = sum_prediction_and_ancestors + len(
            ground_truth_set
        )
    recall = sum_intersection / sum_prediction_and_ancestors
    return recall


def _recall_macro(y_true: np.ndarray, y_pred: np.ndarray):
    return _compute_macro(y_true, y_pred, _recall_micro)


def f1(y_true: np.ndarray, y_pred: np.ndarray, average: str = "micro"):
    r"""
    Compute hierarchical f-score.

    Parameters
    ----------
    y_true : np.array of shape (n_samples, n_levels)
        Ground truth (correct) labels.
    y_pred : np.array of shape (n_samples, n_levels)
        Predicted labels, as returned by a classifier.
    average: {"micro", "macro"}, str, default="micro"
        This parameter determines the type of averaging performed during the computation:

        - `micro`: The f-score is computed by summing over all individual instances, :math:`\displaystyle{hF = \frac{2 \times hP \times hR}{hP + hR}}`, where :math:`hP` is the hierarchical precision and :math:`hR` is the hierarchical recall.
        - `macro`: The f-score is computed for each instance and then averaged, :math:`\displaystyle{hF = \frac{\sum_{i=1}^{n}hF_{i}}{n}}`, where :math:`\alpha_i` is the set consisting of the most specific classes predicted for test example :math:`i` and all their ancestor classes, while :math:`\beta_i` is the set containing the true most specific classes of test example :math:`i` and all their ancestors.
    Returns
    -------
    f1 : float
        Weighted average of the precision and recall
    """
    y_true, y_pred = _validate_input(y_true, y_pred)
    functions = {
        "micro": _f_score_micro,
        "macro": _f_score_macro,
    }
    return functions[average](y_true, y_pred)


def _f_score_micro(y_true: np.ndarray, y_pred: np.ndarray):
    prec = precision(y_true, y_pred)
    rec = recall(y_true, y_pred)
    return 2 * prec * rec / (prec + rec)


def _f_score_macro(y_true: np.ndarray, y_pred: np.ndarray):
    return _compute_macro(y_true, y_pred, _f_score_micro)


def _compute_macro(y_true: np.ndarray, y_pred: np.ndarray, _micro_function):
    overall_sum = 0
    for ground_truth, prediction in zip(y_true, y_pred):
        sample_score = _micro_function(np.array([ground_truth]), np.array([prediction]))
        overall_sum = overall_sum + sample_score
    return overall_sum / len(y_true)



import numpy as np
import pytest
from numpy.testing import assert_array_equal
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils.validation import check_is_fitted

from hiclass import (
    LocalClassifierPerNode,
    LocalClassifierPerLevel,
    LocalClassifierPerParentNode,
)
from hiclass.ConstantClassifier import ConstantClassifier

classifiers = [
    LocalClassifierPerLevel,
    LocalClassifierPerParentNode,
    LocalClassifierPerNode,
]


@pytest.mark.parametrize("classifier", classifiers)
def test_fit_1_class(classifier):
    clf = classifier(local_classifier=LogisticRegression(), n_jobs=2)
    y = np.array([["1", "2"]])
    X = np.array([[1, 2]])
    ground_truth = np.array([["1", "2"]])
    clf.fit(X, y)
    prediction = clf.predict(X)
    assert_array_equal(ground_truth, prediction)


@pytest.fixture
def empty_levels():
    X = [
        [1],
        [2],
        [3],
    ]
    y = np.array(
        [
            ["1"],
            ["2", "2.1"],
            ["3", "3.1", "3.1.2"],
        ],
        dtype=object,
    )
    return X, y


@pytest.mark.parametrize("classifier", classifiers)
def test_empty_levels(empty_levels, classifier):
    clf = classifier()
    X, y = empty_levels
    clf.fit(X, y)
    predictions = clf.predict(X)
    ground_truth = [
        ["1", "", ""],
        ["2", "2.1", ""],
        ["3", "3.1", "3.1.2"],
    ]
    assert list(clf.hierarchy_.nodes) == [
        "1",
        "2",
        "2" + clf.separator_ + "2.1",
        "3",
        "3" + clf.separator_ + "3.1",
        "3" + clf.separator_ + "3.1" + clf.separator_ + "3.1.2",
        clf.root_,
    ]
    assert_array_equal(ground_truth, predictions)


@pytest.mark.parametrize("classifier", classifiers)
def test_fit_bert(classifier):
    bert = ConstantClassifier()
    clf = classifier(
        local_classifier=bert,
        bert=True,
    )
    X = ["Text 1", "Text 2"]
    y = ["a", "a"]
    clf.fit(X, y)
    check_is_fitted(clf)
    predictions = clf.predict(X)
    assert_array_equal(y, predictions)


@pytest.mark.parametrize("classifier", classifiers)
def test_knn(classifier):
    knn = KNeighborsClassifier(
        n_neighbors=2,
    )
    clf = classifier(
        local_classifier=knn,
    )
    y = np.array([["a", "b"], ["a", "c"]])
    X = np.array([[1, 2], [3, 4]])
    clf.fit(X, y)
    check_is_fitted(clf)
    # predictions = lcpn.predict(X)
    # assert_array_equal(y, predictions)


@pytest.mark.parametrize("classifier", classifiers)
def test_fit_multiple_dim_input(classifier):
    clf = classifier()
    X = np.random.rand(1, 275, 3)
    y = np.array([["a", "b", "c"]])
    clf.fit(X, y)
    check_is_fitted(clf)


@pytest.mark.parametrize("classifier", classifiers)
def test_predict_multiple_dim_input(classifier):
    clf = classifier()
    X = np.random.rand(1, 275, 3)
    y = np.array([["a", "b", "c"]])
    clf.fit(X, y)
    predictions = clf.predict(X)
    assert predictions is not None



import numpy as np
import pytest
from pytest import approx

from hiclass.metrics import precision, recall, f1


# TODO: add tests for 3D dataframe (not sure if it's possible to have 3D dataframes)


def test_unmatched_lengths_1d_list():
    y_true = [1, 2, 3]
    y_pred = [1, 2]
    with pytest.raises(AssertionError):
        precision(y_true, y_pred)


def test_unmatched_lengths_2d_list():
    y_true = [[1, 2, 3], [1, 2, 4], [1, 5, 6], [1, 5, 8]]
    y_pred = [[1, 2, 3], [1, 2, 4]]
    with pytest.raises(AssertionError):
        precision(y_true, y_pred)


def test_unmatched_lengths_3d_list():
    y_true = [
        [["human", "mermaid"], ["fish", "mermaid"]],
        [["human", "minotaur"], ["bull", "minotaur"]],
    ]
    y_pred = [
        [["human", "mermaid"], ["fish", "mermaid"]],
    ]
    with pytest.raises(AssertionError):
        precision(y_true, y_pred)


def test_unmatched_lengths_1d_np_array():
    y_true = np.array([1, 2, 3])
    y_pred = np.array([1, 2])
    with pytest.raises(AssertionError):
        precision(y_true, y_pred)


def test_unmatched_lengths_2d_np_array():
    y_true = np.array([[1, 2, 3], [1, 2, 4], [1, 5, 6], [1, 5, 8]])
    y_pred = np.array([[1, 2, 3], [1, 2, 4]])
    with pytest.raises(AssertionError):
        precision(y_true, y_pred)


def test_unmatched_lengths_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
        ]
    )
    with pytest.raises(AssertionError):
        precision(y_true, y_pred)


def test_precision_micro_1d_list():
    y_true = [1, 2, 3, 4]
    y_pred = [1, 2, 5, 6]
    assert 0.5 == precision(y_true, y_pred, "micro")
    assert 1 == precision(y_true, y_true, "micro")


def test_precision_micro_2d_list():
    y_true = [[1, 2, 3, 4], [1, 2, 5, 6]]
    y_pred = [[1, 2, 5, 6], [1, 2, 3, 4]]
    assert 0.5 == precision(y_true, y_pred, "micro")
    assert 1 == precision(y_true, y_true, "micro")


def test_precision_micro_1d_np_array():
    y_true = np.array([1, 2, 3, 4])
    y_pred = np.array([1, 2, 5, 6])
    assert 0.5 == precision(y_true, y_pred, "micro")
    assert 1 == precision(y_true, y_true, "micro")


def test_precision_micro_2d_np_array():
    y_true = np.array([[1, 2, 3, 4], [1, 2, 5, 6]])
    y_pred = np.array([[1, 2, 5, 6], [1, 2, 3, 4]])
    assert 0.5 == precision(y_true, y_pred, "micro")
    assert 1 == precision(y_true, y_true, "micro")


def test_precision_micro_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["", ""]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    assert 0.8333 == approx(precision(y_true, y_pred, "micro"), rel=1e-3)
    assert 1 == precision(y_true, y_true, "micro")


def test_precision_macro_1d_list():
    y_true = [1, 2, 3, 4]
    y_pred = [1, 5, 6, 7]
    assert 0.25 == precision(y_true, y_pred, "macro")
    assert 1 == precision(y_true, y_true, "macro")


def test_precision_macro_2d_list():
    y_true = [[1, 2, 3, 4], [1, 2, 5, 6]]
    y_pred = [[1, 5, 6, 7], [1, 2, 3, 4]]
    assert 0.375 == precision(y_true, y_pred, "macro")
    assert 1 == precision(y_true, y_true, "macro")


def test_precision_macro_1d_np_array():
    y_true = np.array([1, 2, 3, 4])
    y_pred = np.array([1, 5, 6, 7])
    assert 0.25 == precision(y_true, y_pred, "macro")
    assert 1 == precision(y_true, y_true, "macro")


def test_precision_macro_2d_np_array():
    y_true = np.array([[1, 2, 3, 4], [1, 2, 5, 6]])
    y_pred = np.array([[1, 5, 6, 7], [1, 2, 3, 4]])
    assert 0.375 == precision(y_true, y_pred, "macro")
    assert 1 == precision(y_true, y_true, "macro")


def test_precision_macro_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["", ""]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    assert 0.8333 == approx(precision(y_true, y_pred, "macro"), rel=1e-3)
    assert 1 == precision(y_true, y_true, "macro")


def test_recall_micro_1d_list():
    y_true = [1, 2, 3, 4]
    y_pred = [1, 5, 6, 7]
    assert 0.25 == recall(y_true, y_pred, "micro")


def test_recall_micro_2d_list():
    y_true = [[1, 2], [1, 2]]
    y_pred = [[1, 2, 5, 6], [1, 2, 3, 4]]
    assert 1 == recall(y_true, y_pred, "micro")
    assert 0.5 == recall(y_pred, y_true, "micro")


def test_recall_micro_1d_np_array():
    y_true = np.array([1, 2, 3, 4])
    y_pred = np.array([1, 5, 6, 7])
    assert 0.25 == recall(y_true, y_pred, "micro")


def test_recall_micro_2d_np_array():
    y_true = np.array([[1, 2], [1, 2]])
    y_pred = np.array([[1, 2, 5, 6], [1, 2, 3, 4]])
    assert 1 == recall(y_true, y_pred, "micro")
    assert 0.5 == recall(y_pred, y_true, "micro")


def test_recall_micro_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["", ""]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    assert 0.8333 == approx(recall(y_true, y_pred, "micro"), rel=1e-3)
    assert 1 == recall(y_true, y_true, "micro")


def test_recall_macro_1d_list():
    y_true = [1, 2, 3, 5]
    y_pred = [1, 5, 6, 7]
    assert 0.25 == recall(y_true, y_pred, "macro")


def test_recall_macro_2d_list():
    y_true = [[1, 2], [1, 2]]
    y_pred = [[1, 5, 6, 7], [1, 2, 3, 4]]
    assert 0.75 == recall(y_true, y_pred, "macro")
    assert 0.375 == recall(y_pred, y_true, "macro")


def test_recall_macro_1d_np_array():
    y_true = np.array([1, 2, 3, 5])
    y_pred = np.array([1, 5, 6, 7])
    assert 0.25 == recall(y_true, y_pred, "macro")


def test_recall_macro_2d_np_array():
    y_true = np.array([[1, 2], [1, 2]])
    y_pred = np.array([[1, 5, 6, 7], [1, 2, 3, 4]])
    assert 0.75 == recall(y_true, y_pred, "macro")
    assert 0.375 == recall(y_pred, y_true, "macro")


def test_recall_macro_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["", ""]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    assert 0.8333 == approx(recall(y_true, y_pred, "macro"), rel=1e-3)
    assert 1 == recall(y_true, y_true, "macro")


def test_f1_micro_1d_list():
    y_true = [1, 2, 3, 4]
    y_pred = [1, 2, 5, 6]
    assert 0.5 == f1(y_true, y_pred, "micro")


def test_f1_micro_2d_list():
    y_true = [[1, 2, 3, 4], [1, 2, 5, 6]]
    y_pred = [[1, 2, 5, 6], [1, 2, 3, 4]]
    assert 0.5 == f1(y_true, y_pred, "micro")


def test_f1_micro_1d_np_array():
    y_true = np.array([1, 2, 3, 4])
    y_pred = np.array([1, 2, 5, 6])
    assert 0.5 == f1(y_true, y_pred, "micro")


def test_f1_micro_2d_np_array():
    y_true = np.array([[1, 2, 3, 4], [1, 2, 5, 6]])
    y_pred = np.array([[1, 2, 5, 6], [1, 2, 3, 4]])
    assert 0.5 == f1(y_true, y_pred, "micro")


def test_f1_micro_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["", ""]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    assert 0.9090 == approx(f1(y_true, y_pred, "micro"), rel=1e-3)
    assert 1 == f1(y_true, y_true, "micro")


def test_f1_macro_1d_list():
    y_true = [1, 2, 3, 4]
    y_pred = [1, 2, 3, 4]
    assert 1 == f1(y_true, y_pred, "macro")


def test_f1_macro_2d_list():
    y_true = [[1, 2, 3, 4], [1, 2, 5, 6]]
    y_pred = [[1, 5, 6], [1, 2, 3]]
    assert 0.4285714 == approx(f1(y_true, y_pred, "macro"))


def test_f1_macro_1d_np_array():
    y_true = np.array([1, 2, 3, 4])
    y_pred = np.array([1, 2, 3, 4])
    assert 1 == f1(y_true, y_pred, "macro")


def test_f1_macro_2d_np_array():
    y_true = np.array([[1, 2, 3, 4], [1, 2, 5, 6]])
    y_pred = np.array([[1, 5, 6], [1, 2, 3]])
    assert 0.4285714 == approx(f1(y_true, y_pred, "macro"))


def test_f1_macro_3d_np_array():
    y_true = np.array(
        [
            [["human", "mermaid"], ["fish", "mermaid"]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    y_pred = np.array(
        [
            [["human", "mermaid"], ["", ""]],
            [["human", "minotaur"], ["bull", "minotaur"]],
        ]
    )
    assert 0.9 == approx(f1(y_true, y_pred, "macro"), rel=1e-3)
    assert 1 == f1(y_true, y_true, "macro")


def test_empty_levels_2d_list_1():
    y_true = [["2", "3"], ["1"], ["4", "5", "6"]]
    y_pred = [["1"], ["2", "3"], ["4", "5", "6"]]
    assert 0.5 == f1(y_true, y_pred)
    assert 1 == f1(y_true, y_true)


def test_empty_levels_2d_list_2():
    y_true = [["1"], ["2", "3"], ["4", "5", "6"]]
    y_pred = [["1"], ["2", "3"], ["4", "5", "6"]]
    assert 1 == f1(y_true, y_pred)
    assert 1 == f1(y_true, y_true)
